class PageStore:
    def __init__(self, init=None):
        self.pages = init if init else []


    def add(self, endpoint):
        self.pages.append(endpoint)


    def remove(self, endpoint):
        self.pages.remove(endpoint)


    def extend(self, endpoint_list_or_pagestore):
        self.pages.extend(endpoint_list_or_pagestore)


    def extendleft(self, endpoint_list_or_pagestore):
        # TODO probably a way more efficient way to do this using deque or something
        extended = []
        for endpoint in endpoint_list_or_pagestore:
            extended.append(endpoint)

        for endpoint in self.pages:
            extended.append(endpoint)

        self.pages = extended


    def order_by_date(self, descending=True): # sort from latest to earliest (descending) by default
        sort = sorted(self.pages, key=lambda x: x.meta['date'], reverse=descending)
        return PageStore(sort)


    def order_by_title(self, descending=False):
        sort = sorted(self.pages, key=lambda x: x.meta['title'], reverse=descending)
        return PageStore(sort)


    def recent(self):
        return self.order_by_date(descending=True)


    def only_taxonomies(self):
        return PageStore([endpoint for endpoint in self.pages if endpoint.is_taxonomy])


    def exclude_taxonomies(self):
        return PageStore([endpoint for endpoint in self.pages if not endpoint.is_taxonomy])


    def only_dated(self):
        return PageStore([endpoint for endpoint in self.pages if endpoint.meta.get('date')])


    def group_by_date(self):
        # (year, month (1 indexed)): PageStore
        by_months = {}

        for endpoint in self:
            date = endpoint.meta['date']
            by_months.setdefault((date.year, date.month), PageStore()).add(endpoint)

        return by_months


    def filter_by_topic(self, slug):
        filtered = []
        parts = {}
        for endpoint in self.pages:
            if slug in endpoint.taxonomies:
                part = endpoint.taxonomies[slug]
                if part is not None:
                    collision_part = parts.get(part)
                    if collision_part is not None:
                        raise Exception(f"Pages cannot have the same part '{part}' for taxonomy '{slug}': {endpoint.srp} and {collision_part.srp}")
                    parts[part] = endpoint
                filtered.append(endpoint)

        # sort by parts/orders
        filtered_and_sorted = sorted(filtered, key=lambda x: x.taxonomies[slug] if x.taxonomies[slug] is not None else len(filtered))

        #TODO could get rid of the parts = {} and just raise both errors below depending on if part == last_part

        # check to make sure no missing part numbers 1 - X
        last_part = 0
        for endpoint in filtered_and_sorted:
            part = endpoint.taxonomies[slug]
            if part is None:
                continue
            if part != last_part + 1:
                raise Exception(f"Page should be part '{last_part + 1}' for taxonomy '{slug}', not {part}: {endpoint.srp}")
            last_part = part

        return PageStore(filtered_and_sorted)


    def list_view_sort(self, slug):
        # sort precedence:
        # 1. taxonomy order
        # 2. taxonomies in alphabetical order
        # 3. pages in date descending
        # 4. pages with no date last (order doesn't matter)

        ordered = []
        taxonomies = []
        pages = []
        no_date = []

        for endpoint in self.pages:
            tax_order = endpoint.taxonomies.get(slug)
            if tax_order is not None:
                ordered.append(endpoint)
            elif endpoint.is_taxonomy:
                taxonomies.append(endpoint)
            elif endpoint.meta.get('date'):
                pages.append(endpoint)
            else:
                no_date.append(endpoint)

            ordered.sort(key=lambda x: x.taxonomies[slug])
            taxonomies.sort(key=lambda x: x.meta['title'])
            pages.sort(key=lambda x: x.meta['date'], reverse=True)

        return PageStore(ordered + taxonomies + pages + no_date)


    def annotate_nav(self, descending=False):
        recent = self.order_by_date(descending)
        prev = None
        for endpoint in recent:
            if prev:
                prev.next = endpoint
            endpoint.prev = prev
            prev = endpoint

        if recent:
            endpoint.next = None


    def __len__(self):
        return len(self.pages)


    def __getitem__(self, key):
        return self.pages[key]


    def __add__(self, other):
        return PageStore(self.pages + other.pages)



import math

def gen_exp_range(slots, span, reverse=False):
    use_pow = math.log(span, slots)

    use_range = range(slots-1, -1, -1) if reverse else range(slots)

    for i, x in enumerate(use_range):
        yield i, math.floor(math.pow(x+1, use_pow))


def get_links(num_pages, on_page, slots): # on_page 1 indexed
    if num_pages <= slots:
        for x in range(num_pages):
            yield x + 1
        return

    num_pages_left = on_page - 1
    num_pages_right = num_pages - on_page

    available_slots = slots - 1 # the page you're on takes up one # 9
    pages_per_slot = (num_pages - 1) / available_slots # 11

    alloc_slots_left = int(num_pages_left // pages_per_slot) # 2.6 -> 2
    alloc_slots_right = int(num_pages_right // pages_per_slot) # 6.3 -> 6

    # make the allocations whole, favor giving whole one to 'smaller' side

    if alloc_slots_left + alloc_slots_right < available_slots:
        if on_page == 1:
            alloc_slots_right += 1
        elif on_page == num_pages:
            alloc_slots_left += 1
        elif alloc_slots_left < alloc_slots_right:
            alloc_slots_left += 1
        else:
            alloc_slots_right += 1


    if alloc_slots_left == 1:
        yield on_page - 1
    elif alloc_slots_left:
        for i,val in gen_exp_range(alloc_slots_left, num_pages_left, reverse=True):
            use_val = 1 if i == 0 else (on_page - val)
            yield use_val


    yield on_page


    if alloc_slots_right == 1:
        yield on_page + 1 #if on_page < num_pages else num_pages
    elif alloc_slots_right:
        for i,val in gen_exp_range(alloc_slots_right, num_pages_right):
            use_val = num_pages if i == alloc_slots_right - 1 else val + on_page
            yield use_val


class Paginator:

    class Page:
        def __init__(self, paginator, page_num_0, start_index, end_index, prev_endpoint, next_endpoint):
            self.paginator = paginator
            self.items = paginator.items[start_index:end_index]
            self.page_num = page_num_0 + 1 # 1 indexed
            self.start_index = start_index
            self.end_index = end_index
            self.prev_endpoint = prev_endpoint
            self.next_endpoint = next_endpoint

            self.num_items_on_page = len(self.items)

        @property
        def has_next(self):
            return self.next_endpoint is not None

        @property
        def has_prev(self):
            return self.prev_endpoint is not None


    def __init__(self, child_items, per_page):
        self.items = child_items
        self.per_page = per_page
        self.total = len(child_items)
        self.num_of_pages = math.ceil(self.total / per_page)
        self._endpoint_objs = []


    def _attach_page(self, list_endpoint, prev_endpoint): # 0 based page index read off of list_endpoint_obj
        page_num = list_endpoint.page
        start_index = page_num * self.per_page
        end_index = start_index + self.per_page
        if end_index > self.total:
            end_index = self.total

        #prev_endpoint = list_endpoint.prev_page()
        next_endpoint = list_endpoint.next_page() if (page_num + 1) < self.num_of_pages else None

        #return Page(self, page_num, start_index, end_index, prev_endpoint, next_endpoint)
        list_endpoint.paginator_page = self.Page(self, page_num, start_index, end_index, prev_endpoint, next_endpoint)
        self._endpoint_objs.append(list_endpoint)
        return list_endpoint


    def gen_all_pages(self, list_endpoint_0):
        prev_endpoint = None
        annotated_endpoint = self._attach_page(list_endpoint_0, prev_endpoint)
        yield annotated_endpoint

        while annotated_endpoint.paginator_page.next_endpoint is not None:
            prev_endpoint = annotated_endpoint
            next_endpoint = annotated_endpoint.paginator_page.next_endpoint
            annotated_endpoint = self._attach_page(next_endpoint, prev_endpoint)
            yield annotated_endpoint


    def get_page_endpoint(self, page_num): # 1-indexed
        return self._endpoint_objs[page_num - 1]


    def gen_all_page_endpoints(self, on_page, slots): # on_page 1-indexed
        page_num_gen = get_links(self.num_of_pages, on_page, slots) # on_page 1 indexed
        for page_num in page_num_gen:
            yield self.get_page_endpoint(page_num)



from indentgen.default_definitions.config_tag_defs import CONFIG_TAG_SET, SUBSITE_CONFIG_TAG_SET
from indentgen.default_definitions.content_tag_defs import CONTENT_TAG_SET
from indentgen.default_definitions.taxonomy_tag_defs import TAXONOMY_TAG_SET






from uuid import uuid4
import pickle

from indentgen.default_definitions import CONFIG_TAG_SET, SUBSITE_CONFIG_TAG_SET, TAXONOMY_TAG_SET, CONTENT_TAG_SET
import dentmark
from indentgen.img_resize_utils import resize


class PickleableTagDef:
    def __init__(self, context, collectors):
        self.context = context
        self.collectors = collectors


class Wisdom:
    WISDOM_DATA = 'wisdom_data.pickle'
    IMAGE_CACHE_DIR = 'image_cache'
    CACHED_IMAGE_TYPE = 'png'

    def __init__(self, indentgen_inst):
        self.indentgen = indentgen_inst
        self.site_path = indentgen_inst.site_path
        self.wisdom_path = indentgen_inst.wisdom_path
        self.content_path = indentgen_inst.content_path
        self.taxonomy_path = indentgen_inst.taxonomy_path
        self.static_path = indentgen_inst.static_path
        self.img_url = indentgen_inst.IMAGE_URL # the URL prefix where resized images should be accessd from in the HTML
        self.img_output_path = indentgen_inst.img_output_path # directory where resized images should be published
        self.pickle_path = self.wisdom_path / self.WISDOM_DATA

        self.config_file_path = indentgen_inst.config_file_path

        try:
            with open(self.pickle_path, 'rb') as f:
                self.data = pickle.load(f)
        except FileNotFoundError:
            self.data = {}


    def save(self):
        self.wisdom_path.mkdir(parents=True, exist_ok=True)
        with open(self.pickle_path, 'wb') as f:
            pickle.dump(self.data, f)


    def get_static_file_name(self, key_srp):
        abs_static_path = self.static_path / key_srp
        mts = abs_static_path.stat().st_mtime
        static_cache = self.data.get('static_cache')
        if static_cache:
            static_meta = static_cache.get(key_srp)
            if static_meta:
                if static_meta['mts'] == mts:
                    return static_meta['srp']

        static_cache = self.data.setdefault('static_cache', {})
        renamed = f'{key_srp.stem}_{uuid4().hex}{key_srp.suffix}'
        new_srp = key_srp.with_name(renamed)
        static_cache[key_srp] = {'mts': mts, 'srp': new_srp}
        self.save()
        return new_srp


    def get_rendered(self, key_srp, is_taxonomy = False, meta_only=False):
        abs_content_path = self.site_path / key_srp
        abs_cached_path = (self.wisdom_path / key_srp).with_suffix('.html')

        mts = abs_content_path.stat().st_mtime

        if not meta_only:
            render_cache = self.data.get('render_cache')
            if render_cache:
                render_meta = render_cache.get(key_srp)
                if render_meta:
                    if render_meta['mts'] == mts:
                        try:
                            with open(abs_cached_path, 'r') as f:
                                root = render_meta['root']
                                return f.read(), root
                        except FileNotFoundError:
                            # CACHED render result doesn't exist - re-render
                            pass

        extra_context = {
            'srp': key_srp,
            'indentgen': self.indentgen
        }

        with open(abs_content_path, 'r') as f:
            try:
                only_address = 'root.meta' if meta_only else None
                root = dentmark.parse(f, TAXONOMY_TAG_SET if is_taxonomy else CONTENT_TAG_SET, extra_context, only_address)
            except Exception as e:
                raise Exception(f'{abs_content_path}: {e}')

        root.pre_render()

        if meta_only:
            return '', root

        try:
            rendered = root.render()
        except Exception as e:
            raise Exception(f'{abs_content_path}: {e}')

        abs_cached_path.parent.mkdir(parents=True, exist_ok=True)
        with open(abs_cached_path, 'w') as f:
            f.write(rendered)

        render_cache = self.data.setdefault('render_cache', {})
        pickleable_root = PickleableTagDef(root.context, root.collectors)
        render_cache[key_srp] = {'mts': mts, 'root': pickleable_root}

        self.save()
        return rendered, root


    def get_config(self):
        mts = self.config_file_path.stat().st_mtime

        config_cache = self.data.get('config_cache')
        if config_cache:
            config_mts = config_cache['mts']
            if config_mts == mts:
                return config_cache['data']

        with open(self.config_file_path, 'r') as f:
            try:
                config_data = dentmark.render(f, CONFIG_TAG_SET)
            except Exception as e:
                raise Exception(f'{self.config_file_path}: {e}')

        self.data['config_cache'] = {'mts': mts, 'data': config_data}

        self.save()

        return config_data


    def get_subsite_config(self, subsite_config_dir_srp):
        config_file_path = self.site_path / subsite_config_dir_srp / self.indentgen.CONFIG_FILE_NAME
        mts = config_file_path.stat().st_mtime

        subsite_cache = self.data.get('subsite_cache')
        if subsite_cache:
            subsite_data = subsite_cache.get(subsite_config_dir_srp)
            if subsite_data:
                config_mts = subsite_data['mts']
                if config_mts == mts:
                    return subsite_data['data']

        with open(config_file_path, 'r') as f:
            try:
                config_data = dentmark.render(f, SUBSITE_CONFIG_TAG_SET)
            except Exception as e:
                raise Exception(f'{config_file_path}: {e}')

        self.data.setdefault('subsite_cache', {})[subsite_config_dir_srp] = {'mts': mts, 'data': config_data}
        self.save()
        return config_data


    def _get_or_create_image_version(self, relative_path, max_width, max_height, copy_original, dentmark_srp):

        cache_key = (relative_path, max_width, max_height)

        img_cache = self.data.get('img_cache')
        if img_cache:
            img_data = img_cache.get(cache_key)
            if img_data:
                img_data['copy_original'] = copy_original
                return relative_path, img_data['serve_path'], img_data['original_serve_path']


        full_path = self.site_path / relative_path

        mts = full_path.stat().st_mtime

        parent = relative_path.parent

        wisdom_save_dir = self.wisdom_path / self.IMAGE_CACHE_DIR / parent
        original_name_stem = full_path.stem

        new_filename = f'{original_name_stem}_{max_width}_{max_height}.{self.CACHED_IMAGE_TYPE}'
        cached_img_path = wisdom_save_dir / new_filename
        publish_path = (self.img_output_path / relative_path).with_name(new_filename)
        serve_path = ('/' + self.img_url) / publish_path.relative_to(self.img_output_path)
        original_publish_path = (self.img_output_path / relative_path)
        original_serve_path = ('/' + self.img_url) / original_publish_path.relative_to(self.img_output_path)

        img_cache = self.data.setdefault('img_cache', {})

        img_cache[cache_key] = {
            'mts': mts,
            'cached_path': cached_img_path,
            'publish_path': publish_path,
            'serve_path': serve_path,
            'original_path': full_path,
            'max_width': max_width,
            'max_height': max_height,
            'relative_path': relative_path,
            'copy_original': copy_original,
            'original_serve_path': original_serve_path,
            'original_publish_path': original_publish_path
        }

        # This is only needed for generating helpful exception messages. If this is called via get_image_url_by_key,
        # then it's being called by a template and will have already been checked so it's safe to leave this out in that case,
        # since the template has no way of passing this value
        if dentmark_srp:
            img_cache[cache_key]['dentmark_srp'] = dentmark_srp

        self.save()

        return relative_path, serve_path, original_serve_path


    def get_image_url(self, dentmark_srp, image_relative_path, max_width, max_height, copy_original=False): # used by dentmark
        parent = dentmark_srp.parent
        full_path = self.site_path / parent / image_relative_path

        try:
            resolved = full_path.resolve(True) # make sure the image exists
            relative = resolved.relative_to(self.site_path) # make sure relative path doesn't ascend past the site_path
        except (ValueError, FileNotFoundError) as e:
            raise Exception(f"Invalid image url '{image_relative_path}'")

        return self._get_or_create_image_version(relative, max_width, max_height, copy_original, dentmark_srp)


    def get_image_url_by_key(self, srp_key, max_width, max_height): # used by templates
        return self._get_or_create_image_version(srp_key, max_width, max_height, False, None)


    def gen_cached_images(self, build=False):
        for img_data in self.data.get('img_cache', {}).values():
            if build:
                original_path = img_data['original_path']
                cached_path = img_data['cached_path']

                # TODO will prob. error if original image no longer exists. Catch and handle better
                try:
                    mts = original_path.stat().st_mtime
                except FileNotFoundError as e:
                    #raise Exception(f"Error in {img_data['dentmark_srp']}: Image file not found: {img_data['relative_path']}")
                    print(f"Stale Cached Image in Wisdom. No longer exists in content: {img_data['relative_path']}")
                    continue

                if mts != img_data['mts'] or not cached_path.exists():
                    print(f"Resizing image to {img_data['max_width']}x{img_data['max_height']}: {original_path}")
                    resize(original_path, cached_path, img_data['max_width'], img_data['max_height'])
                    img_data['mts'] = mts
                    self.save()

            yield img_data



import re
from dentmark import defs_manager, TagDef, OptionalUnique, RequiredUnique
from dentmark.default_definitions import Root
from indentgen.default_definitions.content_tag_defs import MetaContext, TitleMetaContext
from indentgen.taxonomy_def_set import MetaTaxonomyContext

TAXONOMY_TAG_SET = 'indentgen_taxonomy'

taxonomy_tag_set = defs_manager.copy_tag_set(TAXONOMY_TAG_SET)


@taxonomy_tag_set.register(replace=True)
class IndentgenTaxonomyRoot(Root):

    def before_render(self):
        meta = self.context['meta']
        taxonomies = meta.get('taxonomy', {})
        slug_path = meta['slug_path']

        slug_path_components = slug_path.split('/')
        parent_slug_path = '/'.join(slug_path_components[:-1])

        if parent_slug_path and parent_slug_path not in taxonomies:
            taxonomies = meta.setdefault('taxonomy', {})
            taxonomies[parent_slug_path] = None

        if slug_path in taxonomies:
            return f"Cannot list self as a taxonomy: '{tax_slug_path}'"


@taxonomy_tag_set.register()
class TaxonomyPseudoContext(TagDef):
    tag_name = 'pseudo'
    is_context = True

    min_num_text_nodes = 0
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.meta')]

    def process_data(self, data):
        if data:
            return data[0].lower() == 'true'
        return True

    def validate(self):
        if self.children:
            if self.children[0].get_data().lower() not in ('true', 'false'):
                return f"'pseudo' tags value must be either 'true', 'false', or [empty]. Defaults to 'true' if [empty]"


@taxonomy_tag_set.register()
class TaxonomyGalleryContext(TaxonomyPseudoContext):
    tag_name = 'gallery'

    def validate(self):
        if self.children:
            if self.children[0].get_data().lower() not in ('true', 'false'):
                return f"'gallery' tags value must be either 'true', 'false', or [empty]. Defaults to 'true' if [empty]"


@taxonomy_tag_set.register()
class ConfigSlugPathContext(TagDef):
    tag_name = 'slug_path'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [RequiredUnique('root.meta')]

    url_pattern = re.compile(r'^[a-z0-9](?:[a-z0-9-]*[a-z0-9])?(?:/[a-z0-9](?:[a-z0-9-]*[a-z0-9])?)*$')

    def validate(self):
        val = self.get_data()
        if not self.url_pattern.match(val):
            return f"Tag '{self.tag_name}' contains invalid format for a slug. Slug components can contain only [a-z0-9-] and cannot begin or end with '-' and must be separated by '/' with no leading and trailing '/'"


taxonomy_tag_set.register_tag(MetaContext)
taxonomy_tag_set.register_tag(MetaTaxonomyContext)
taxonomy_tag_set.register_tag(TitleMetaContext)



from PIL import Image

def resize(abs_path_in, abs_path_out, max_width, max_height):
    with Image.open(abs_path_in) as im:
        if im.width > max_width or im.height > max_height:
            im.thumbnail((max_width, max_height))

        abs_path_out.parent.mkdir(parents=True, exist_ok=True)
        im.save(abs_path_out, "PNG")





from pathlib import Path

class PathDict:
    def __init__(self):
        self._ordered_path_keys = []
        self._data_dict = {}


    def add(self, path_or_f, data):
        key = Path(path_or_f)

        if key not in self._data_dict:
            # key does not exist. Add new key and order it
            self._ordered_path_keys.append(key)
            self._ordered_path_keys.sort(key=lambda x: len(x.parts), reverse=True)

        self._data_dict[key] = data


    def get(self, path_or_f):
        # fetches the key that matches the closest relative path
        key = Path(path_or_f)

        for path in self._ordered_path_keys:
            try:
                key.relative_to(path)
                return self._data_dict[path]
            except ValueError:
                pass
        return None


    def __iter__(self):
        # when iterating over these to validate, we should check the highest path first (i.e. path with least parts)
        # in case children subsites depend on a parent higher in the path tree. Since the keys are ordered from paths with the most
        # parts to paths with the least, iterate over this object in reverse key order.
        for key in reversed(self._ordered_path_keys):
            yield self._data_dict[key]
